def merge_dict(*args, add_keys=True):
    """
    Safely merge two dictionaries

    Args:
        dct0...n: dicts to merge
        add_keys: merge dict keys (default: True)
    Returns:
        merged dict
    """
    if len(args) < 1: return None
    dct = args[0].copy()
    from collections.abc import Mapping
    for merged in args[1:]:
        if not add_keys:
            merged = {k: merged[k] for k in set(dct).intersection(set(merged))}

        for k, v in merged.items():
            if isinstance(dct.get(k), dict) and isinstance(v, Mapping):
                dct[k] = merge_dict(dct[k], v, add_keys=add_keys)
            else:
                if v is None:
                    if not k in dct:
                        dct[k] = None
                else:
                    dct[k] = v
    return dct


def val_to_boolean(val):
    """
    Convert any value to boolean

    Boolean: return as-is

    Integer: 1 = True, 0 = False

    String (case-insensitive): 1, 'true', 'yes', 'on', 'y' = True
                               0, 'false', 'no', 'off', 'n' = False

    Args:
        val: value to convert

    Returns:
        boolean converted value, None if val is None

    Raises:
        ValueError: if value can not be converted
    """
    if val is None: return None
    elif isinstance(val, bool): return val
    else:
        val = str(val)
        if val.lower() in ['1', 'true', 'yes', 'on', 'y']: return True
        elif val.lower() in ['0', 'false', 'no', 'off', 'n']: return False
        else: raise ValueError


def safe_int(val):
    """
    Convert string/float to integer

    If input value is integer - return as-is
    If input value is a hexadecimal (0x00): converts hex to decimal

    Args:
        val: value to convert

    Raises:
        ValueError: if input value can not be converted
    """
    if isinstance(val, int):
        return val
    elif isinstance(val, str) and val.find('x') != -1:
        return int(val, 16)
    else:
        return int(val)


def parse_date(val=None, return_timestamp=True):
    """
    Parse date from string or float/integer

    Input date can be either timestamp or date-time string

    If input value is integer and greater than 3000, it's considered as a
    timestamp, otherwise - as a year

    Args:
        val: value to parse
        return_timestamp: return UNIX timestamp (default) or datetime object

    Returns:
        UNIX timestamp (float) or datetime object. If input value is None,
        returns current date/time
    """
    import datetime
    import time
    if val is None:
        return time.time() if return_timestamp else datetime.datetime.now()
    if isinstance(val, datetime.datetime):
        dt = val
    else:
        try:
            val = float(val)
            if val > 3000:
                return val if return_timestamp else \
                        datetime.datetime.fromtimestamp(val)
            else:
                val = int(val)
        except:
            pass
        import dateutil.parser
        dt = dateutil.parser.parse(str(val))
    return dt.timestamp() if return_timestamp else dt


def parse_number(val):
    """
    Tries to parse number from any value

    Valid values are:

        any float / integer
        123.45
        123 456.899
        123,456.899
        123 456,899
        123.456,82

    Args:
        val: value to parse
    Returns:
        val as-is if val is integer, float or None, otherwise parsed value
    Raises:
        ValueError: if input val can not be parsed
    """
    if isinstance(val, int) or isinstance(val, float) or val is None:
        return val
    if not isinstance(val, str):
        raise ValueError(val)
    try:
        return float(val)
    except:
        pass
    spaces = val.count(' ')
    commas = val.count(',')
    dots = val.count('.')
    if spaces > 0:
        return float(val.replace(' ', '').replace(',', '.'))
    elif commas > 1:
        return float(val.replace(',', ''))
    elif commas == 1 and commas <= dots:
        if val.find(',') < val.find('.'):
            return float(val.replace(',', ''))
        else:
            return float(val.replace('.', '').replace(',', '.'))
    else:
        return float(val.replace(',', '.'))



import re
function_wrong_symbols = re.compile(r"[<>{}[\]~`]")


def parse_func_str(s):
    """
    Parse string to a function

    Parses a given string to a Python-like function

    Examples of valid strings:
    
        myfunc()
        myfunc('test', '123')
        myfunc(test, 123)
        myfunc(1,2,3)
        myfunc(1,2,name='test')
        myfunc(name=test)
        myfunc(name='test',value='123')

    Returns:
        tuple: (function_name, [list of args], {dict of kwargs})
    Raises:
        ValueError
    """
    s = s.strip()
    if not s.endswith(')'):
        raise ValueError('ERROR: argument doesn\'t have brackets')
    r = s.replace(')', '').split('(')
    name = r.pop(0).strip()
    if function_wrong_symbols.search(name) or name.find(' ') != -1:
        raise ValueError('Invalid symbols in argument')
    args = []
    list_args = []
    list_kw = []
    kw = {}
    check = re.compile(r'(,*\s*\'*\w+\'*\s*[=]\s*[\'\"]*.*[\'|\"]*)')
    argum = [i.strip() for i in check.split(r[0]) if i]
    for a in argum:
        a = a.replace(',', '', 1).strip() if a.startswith(',') else a
        if not a.__contains__('='):
            list_args = a
        elif a.__contains__('='):
            if a.startswith('\'') or a.startswith('\"'):
                list_args = a
            else:
                list_kw = a
    if list_args:
        clear_arg = [
            a.strip()
            for a in re.split(
                r'(,*\s*\w*\s*),|(\s*[\"|\'][\w\s,]*[\s\w]*[\"|\'])', list_args)
            if a
        ]
        for t in clear_arg:
            t = t.replace(',', '').strip() if (t.startswith(',') or
                                               t.endswith(',')) else t
            if t.startswith('\'') or t.startswith('"'):
                args.append(t)
            else:
                try:
                    args.append(int(t))
                except ValueError:
                    try:
                        args.append(float(t))
                    except ValueError:
                        args.append(t)
    if list_kw:
        clear_kw = [
            k.strip() for k in re.split(
                r'([,*\s\w]*[=]\s*[\'|\"]*[\w\s=\'\"%]*[,\s\w]*[\'|\",])',
                list_kw) if k
        ]
        for t in clear_kw:
            t = t.replace(',', '').strip() if (t.startswith(',') or
                                               t.endswith(',')) else t
            if not t:
                continue
            k, v = [x.strip() for x in t.split('=', 1) if t]
            if function_wrong_symbols.search(k) or k.find(' ') != -1:
                raise ValueError('Invalid kwargs param')
            try:
                kw[k] = int(v)
            except ValueError:
                try:
                    kw[k] = float(v)
                except ValueError:
                    kw[k] = ''.join(list(v)[1:-1]) if \
                            ((v.__contains__('\'') or v.__contains__('"')) and \
                             list(v)[0] == list(v)[-1]) else v
    return name, args, kw



def gen_random_str(length=32):
    """
    Generate random string (letters+numbers)

    Args:
        length: string length (default: 32)
    """
    import string
    import random
    symbols = string.ascii_letters + '0123456789'
    return ''.join(random.choice(symbols) for i in range(length))


def _pad_key(key):
    l = len(key)
    if l == 16 or l == 24 or l == 32:
        return key
    elif l < 16:
        return key.ljust(16)
    elif l < 24:
        return key.ljust(24)
    elif l < 32:
        return key.ljust(32)
    else:
        return key[:32]


def encrypt(raw, key, encode=True):
    """
    Encrypt bytes with AES-CBC

    For AES-128: key size <= 16 bytes
    For AES-192: key size <= 24 bytes
    For AES-256: key size <= 32 bytes

    Keys, longer than 32 bytes, are truncated

    Args:
        raw: bytes to encrypt
        key: encryption key
        encode: encode result in base64 (default: True)
    """
    from Crypto.Cipher import AES
    from Crypto import Random
    length = 16 - (len(raw) % 16)
    raw += bytes([length]) * length
    iv = Random.new().read(AES.block_size)
    cipher = AES.new(_pad_key(key), AES.MODE_CBC, iv)
    val = iv + cipher.encrypt(raw)
    if encode:
        import base64
        return base64.b64encode(val).decode()
    else:
        return val


def decrypt(enc, key, decode=True):
    """
    Decrypt encoded data with AES-CBC

    Args:
        enc: data to decrypt
        key: decryption key
        decode: decode data from base64 (default: True)
    """
    from Crypto.Cipher import AES
    if decode:
        import base64
        enc = base64.b64decode(enc)
    iv = enc[:16]
    cipher = AES.new(_pad_key(key), AES.MODE_CBC, iv)
    data = cipher.decrypt(enc[16:])
    return data[:-data[-1]]



"""
JSON processing and output functions

Automatically imports rapidjson if present
"""
import sys

try:
    import rapidjson as json
except:
    import json


def loads(raw, **kwargs):
    """
    Load JSON

    Args:
        raw: raw data to parse
        **kwargs: sent to json.loads() as-is
    """
    return json.loads(raw)


def dumps(data, pretty=False, unpicklable=False, pickle_opts={}, **kwargs):
    """
    Dump to JSON

    Args:
        data: data to dump
        pretty: set indent and sort keys
        unpicklable: one-way dump for complex objects (requires jsonpickle)
        pickle_opts: sent to jsonpickle.encode() as-is
        **kwargs: sent to json.dumps() as-is
    """
    if unpicklable:
        import jsonpickle
        data = json.loads(
            jsonpickle.encode(data, unpicklable=True, **pickle_opts))
    return json.dumps(data, indent=4, sort_keys=True, **
                      kwargs) if pretty else json.dumps(data, **kwargs)


def jprint(data, colored=True, force_colored=False, file=None):
    """
    Pretty print JSON

    Args:
        data: data to encode and print
        colored: colorize output (default: True)
        force_colored: force colorize, even if stream is not a tty
        file: output stream (default: sys.stdout)
    """
    j = dumps(data, pretty=True, unpicklable=True)
    if file is None: file = sys.stdout
    if colored and (file.isatty() or force_colored):
        from pygments import highlight, lexers, formatters
        j = highlight(j, lexers.JsonLexer(), formatters.TerminalFormatter())
    print(j, file=file)
